import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import org.tidestudio.utils.TideIOUtils;
import org.tidestudio.utils.TideFSUtils;

import weka.attributeSelection.Evaluator;
import weka.attributeSelection.PSOSearch;
import weka.classifiers.lazy.IBk;
import weka.core.Instances;

public final class FeatureSelection implements Runnable {

    public class TestJob {
        public String  JobType;
        public boolean saveParam;
        public boolean saveResult;

        public TestJob(final String jt, final boolean sp, final boolean sr) {
            JobType = jt;
            saveParam = sp;
            saveResult = sr;
        }
    }

    private List<TestJob> m_jobList;

    private final String  m_dataFile;
    private final String  m_logFile;
    private final String  m_objFile;

    private long          m_randSeedShuffle;
    private long          m_randSeedAlgorithm;

    private double        m_trainPortion;
    private double        m_optiPortion;

    private int           m_nInstances, m_nOpti, m_nTrain;

    // Cross Validation
    private int           m_nFoldCV;

    // Knn
    private IBk           m_ibk;
    private int           m_kMax, m_kOpted;
    private boolean       m_optimizeK;
    private double        m_errorKnn;
    private Evaluator     m_eva;

    // PSO Configuration
    private int           m_popSize;
    private int           m_maxIter;
    private double        m_psoW;
    private double        m_psoC1;
    private double        m_psoC2;
    private double        m_psoPhi;

    // Init Configuration
    private String        m_initTypeString;
    private double[]      m_initPortion;
    private double[]      m_initProb;

    private PSOSearch     m_pso;

    private int[]         m_selectionPSO, m_selectionCSO;
    private double        m_errorPSO, m_errorCSO;

    private Instances     m_all, m_opti, m_train, m_test;

    private DecimalFormat df2 = new DecimalFormat("#.##");

    public FeatureSelection(final String dataset, final String logfile) throws Exception {
        this(dataset, logfile, System.nanoTime());
    }

    public FeatureSelection(final String dataset, final String logfile, long randSeedShuffle) throws Exception {
        m_dataFile = dataset;
        m_logFile = logfile;
        m_nFoldCV = 10;
        m_randSeedShuffle = randSeedShuffle;

        StringBuffer sb = new StringBuffer();
        sb.append(dataset.substring(0, dataset.lastIndexOf('/') + 1));
        sb.append(m_dataFile.substring(m_dataFile.lastIndexOf('/') + 1, m_dataFile.lastIndexOf('.')));
        sb.append('-').append(m_randSeedShuffle).append(".ser");
        m_objFile = sb.toString();

        resetOptions();
        readData();
        prepData();
        initKnn();
    }

    @Override
    public void run() {
        try {
            for (TestJob job : m_jobList) {
                switch (job.JobType) {
                    case "PSO":
                        testPSO(job.saveParam, job.saveResult);
                        break;
                    case "CSO":
                        testCSO(job.saveParam, job.saveResult);
                        break;
                    case "Xue1":
                        testXue1(job.saveParam, job.saveResult);
                        break;
                    case "Xue2":
                        testXue2(job.saveParam, job.saveResult);
                        break;
                    case "Xue3":
                        testXue3(job.saveParam, job.saveResult);
                        break;
                    case "Xue4":
                        testXue4(job.saveParam, job.saveResult);
                        break;
                    default:
                        assert (false);
                        System.err.println("Incorrect Job Type!");
                        break;
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        MainClass.completed();
    }

    public void addJob(final String jobType, final boolean saveParam, final boolean saveResult) {
        if (m_jobList == null) {
            m_jobList = new ArrayList<TestJob>();
        }
        m_jobList.add(new TestJob(jobType, saveParam, saveResult));
    }

    public void testXue1(final boolean saveParam, final boolean saveResult) throws Exception {
        if (saveParam) {
            logParameters("Xue1");
        }
        m_pso.setPopulationSize(m_popSize);
        m_pso.setIterations(m_maxIter);
        m_pso.setC1(m_psoC1);
        m_pso.setC2(m_psoC2);
        m_pso.setW(m_psoW);
        m_pso.setPhi(m_psoPhi);
        m_pso.setThreshold(0.6);
        m_pso.setInstPortionProb(new double[] { 1. }, new double[] { 0.5 });
        m_pso.setSeed(m_randSeedAlgorithm);

        m_pso.setLookupTableFilePath(m_objFile);

        m_selectionPSO = m_pso.search(m_eva, m_opti, 0);

        Instances train = TideFSUtils.instancesBySelectedFeatures(m_train, m_selectionPSO);
        Instances test = TideFSUtils.instancesBySelectedFeatures(m_test, m_selectionPSO);
        m_ibk.buildClassifier(train);
        m_errorPSO = TideFSUtils.errorRate(m_ibk, test);
        if (saveResult) {
            TideIOUtils.saveLog(m_logFile, "Number of evaluations: " + TideIOUtils.arrayToString(m_pso.getNumEvaluation()));
            TideIOUtils.saveLog(m_logFile, "Hist fitness: " + TideIOUtils.arrayToString(m_pso.getHistFitness()));
            TideIOUtils.saveLog(m_logFile, "Hist nFeatures: " + TideIOUtils.arrayToString(m_pso.getHistNumSelected()));
            TideIOUtils.saveLog(m_logFile, "Hist Selection: " + TideIOUtils.listListToString(m_pso.getHistSelected()));
            TideIOUtils.saveLog(m_logFile, "Selection: " + TideIOUtils.arrayToString(m_selectionPSO));
            TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN Xue1: " + m_errorPSO);
            TideIOUtils.saveLog(m_logFile, "");
        }
    }

    public void testXue2(final boolean saveParam, final boolean saveResult) throws Exception {
        if (saveParam) {
            logParameters("Xue2");
        }
        m_pso.setPopulationSize(m_popSize);
        m_pso.setIterations(m_maxIter);
        m_pso.setC1(m_psoC1);
        m_pso.setC2(m_psoC2);
        m_pso.setW(m_psoW);
        m_pso.setPhi(m_psoPhi);
        m_pso.setThreshold(0.6);
        m_pso.setInstPortionProb(new double[] { 1. }, new double[] { 0.1 });
        m_pso.setSeed(m_randSeedAlgorithm);

        m_pso.setLookupTableFilePath(m_objFile);

        m_selectionPSO = m_pso.search(m_eva, m_opti, 0);

        Instances train = TideFSUtils.instancesBySelectedFeatures(m_train, m_selectionPSO);
        Instances test = TideFSUtils.instancesBySelectedFeatures(m_test, m_selectionPSO);
        m_ibk.buildClassifier(train);
        m_errorPSO = TideFSUtils.errorRate(m_ibk, test);
        if (saveResult) {
            TideIOUtils.saveLog(m_logFile, "Number of evaluations: " + TideIOUtils.arrayToString(m_pso.getNumEvaluation()));
            TideIOUtils.saveLog(m_logFile, "Hist fitness: " + TideIOUtils.arrayToString(m_pso.getHistFitness()));
            TideIOUtils.saveLog(m_logFile, "Hist nFeatures: " + TideIOUtils.arrayToString(m_pso.getHistNumSelected()));
            TideIOUtils.saveLog(m_logFile, "Hist Selection: " + TideIOUtils.listListToString(m_pso.getHistSelected()));
            TideIOUtils.saveLog(m_logFile, "Selection: " + TideIOUtils.arrayToString(m_selectionPSO));
            TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN Xue2: " + m_errorPSO);
            TideIOUtils.saveLog(m_logFile, "");
        }
    }

    public void testXue3(final boolean saveParam, final boolean saveResult) throws Exception {
        if (saveParam) {
            logParameters("Xue3");
        }
        m_pso.setPopulationSize(m_popSize);
        m_pso.setIterations(m_maxIter);
        m_pso.setC1(m_psoC1);
        m_pso.setC2(m_psoC2);
        m_pso.setW(m_psoW);
        m_pso.setPhi(m_psoPhi);
        m_pso.setThreshold(0.6);
        m_pso.setInstPortionProb(new double[] { 1. }, new double[] { 2. / 3 });
        m_pso.setSeed(m_randSeedAlgorithm);

        m_pso.setLookupTableFilePath(m_objFile);

        m_selectionPSO = m_pso.search(m_eva, m_opti, 0);

        Instances train = TideFSUtils.instancesBySelectedFeatures(m_train, m_selectionPSO);
        Instances test = TideFSUtils.instancesBySelectedFeatures(m_test, m_selectionPSO);
        m_ibk.buildClassifier(train);
        m_errorPSO = TideFSUtils.errorRate(m_ibk, test);
        if (saveResult) {
            TideIOUtils.saveLog(m_logFile, "Number of evaluations: " + TideIOUtils.arrayToString(m_pso.getNumEvaluation()));
            TideIOUtils.saveLog(m_logFile, "Hist fitness: " + TideIOUtils.arrayToString(m_pso.getHistFitness()));
            TideIOUtils.saveLog(m_logFile, "Hist nFeatures: " + TideIOUtils.arrayToString(m_pso.getHistNumSelected()));
            TideIOUtils.saveLog(m_logFile, "Hist Selection: " + TideIOUtils.listListToString(m_pso.getHistSelected()));
            TideIOUtils.saveLog(m_logFile, "Selection: " + TideIOUtils.arrayToString(m_selectionPSO));
            TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN Xue3: " + m_errorPSO);
            TideIOUtils.saveLog(m_logFile, "");
        }
    }

    public void testXue4(final boolean saveParam, final boolean saveResult) throws Exception {
        if (saveParam) {
            logParameters("Xue4");
        }
        m_pso.setPopulationSize(m_popSize);
        m_pso.setIterations(m_maxIter);
        m_pso.setC1(m_psoC1);
        m_pso.setC2(m_psoC2);
        m_pso.setW(m_psoW);
        m_pso.setPhi(m_psoPhi);
        m_pso.setThreshold(0.6);
        m_pso.setInstPortionProb(new double[] { 2. / 3, 1. / 3 }, new double[] { 0.1, 2. / 3 });
        m_pso.setSeed(m_randSeedAlgorithm);

        m_pso.setLookupTableFilePath(m_objFile);

        m_selectionPSO = m_pso.search(m_eva, m_opti, 0);

        Instances train = TideFSUtils.instancesBySelectedFeatures(m_train, m_selectionPSO);
        Instances test = TideFSUtils.instancesBySelectedFeatures(m_test, m_selectionPSO);
        m_ibk.buildClassifier(train);
        m_errorPSO = TideFSUtils.errorRate(m_ibk, test);
        if (saveResult) {
            TideIOUtils.saveLog(m_logFile, "Number of evaluations: " + TideIOUtils.arrayToString(m_pso.getNumEvaluation()));
            TideIOUtils.saveLog(m_logFile, "Hist fitness: " + TideIOUtils.arrayToString(m_pso.getHistFitness()));
            TideIOUtils.saveLog(m_logFile, "Hist nFeatures: " + TideIOUtils.arrayToString(m_pso.getHistNumSelected()));
            TideIOUtils.saveLog(m_logFile, "Hist Selection: " + TideIOUtils.listListToString(m_pso.getHistSelected()));
            TideIOUtils.saveLog(m_logFile, "Selection: " + TideIOUtils.arrayToString(m_selectionPSO));
            TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN Xue4: " + m_errorPSO);
            TideIOUtils.saveLog(m_logFile, "");
        }
    }

    public void testPSO(final boolean saveParam, final boolean saveResult) throws Exception {
        if (saveParam) {
            logParameters("PSO");
        }
        m_pso.setPopulationSize(m_popSize);
        m_pso.setIterations(m_maxIter);
        m_pso.setC1(m_psoC1);
        m_pso.setC2(m_psoC2);
        m_pso.setW(m_psoW);
        m_pso.setPhi(m_psoPhi);
        m_pso.setInstPortionProb(m_initPortion, m_initProb);
        m_pso.setSeed(m_randSeedAlgorithm);

        m_pso.setLookupTableFilePath(m_objFile);

        m_selectionPSO = m_pso.search(m_eva, m_opti, 0);

        Instances train = TideFSUtils.instancesBySelectedFeatures(m_train, m_selectionPSO);
        Instances test = TideFSUtils.instancesBySelectedFeatures(m_test, m_selectionPSO);
        m_ibk.buildClassifier(train);
        m_errorPSO = TideFSUtils.errorRate(m_ibk, test);
        if (saveResult) {
            TideIOUtils.saveLog(m_logFile, "Number of evaluations: " + TideIOUtils.arrayToString(m_pso.getNumEvaluation()));
            TideIOUtils.saveLog(m_logFile, "Hist fitness: " + TideIOUtils.arrayToString(m_pso.getHistFitness()));
            TideIOUtils.saveLog(m_logFile, "Hist nFeatures: " + TideIOUtils.arrayToString(m_pso.getHistNumSelected()));
            TideIOUtils.saveLog(m_logFile, "Hist Selection: " + TideIOUtils.listListToString(m_pso.getHistSelected()));
            TideIOUtils.saveLog(m_logFile, "Selection: " + TideIOUtils.arrayToString(m_selectionPSO));
            TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN PSO: " + m_errorPSO);
            TideIOUtils.saveLog(m_logFile, "");
        }
    }

    public void testCSO(final boolean saveParam, final boolean saveResult) throws Exception {
        if (saveParam) {
            logParameters("CSO");
        }
        m_pso.setPopulationSize(m_popSize);
        m_pso.setIterations(m_maxIter * 2);
        m_pso.setC1(m_psoC1);
        m_pso.setC2(m_psoC2);
        m_pso.setW(m_psoW);
        m_pso.setPhi(m_psoPhi);
        m_pso.setInstPortionProb(m_initPortion, m_initProb);
        m_pso.setSeed(m_randSeedAlgorithm);

        m_pso.setLookupTableFilePath(m_objFile);

        m_selectionCSO = m_pso.search(m_eva, m_opti, 1);

        Instances train = TideFSUtils.instancesBySelectedFeatures(m_train, m_selectionCSO);
        Instances test = TideFSUtils.instancesBySelectedFeatures(m_test, m_selectionCSO);
        m_ibk.buildClassifier(train);
        m_errorCSO = TideFSUtils.errorRate(m_ibk, test);
        if (saveResult) {
            TideIOUtils.saveLog(m_logFile, "Number of evaluations: " + TideIOUtils.arrayToString(m_pso.getNumEvaluation()));
            TideIOUtils.saveLog(m_logFile, "Hist fitness: " + TideIOUtils.arrayToString(m_pso.getHistFitness()));
            TideIOUtils.saveLog(m_logFile, "Hist nFeatures: " + TideIOUtils.arrayToString(m_pso.getHistNumSelected()));
            TideIOUtils.saveLog(m_logFile, "Hist Selection: " + TideIOUtils.listListToString(m_pso.getHistSelected()));
            TideIOUtils.saveLog(m_logFile, "Selection: " + TideIOUtils.arrayToString(m_selectionCSO));
            TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN CSO: " + m_errorCSO);
            TideIOUtils.saveLog(m_logFile, "");
        }
    }

    public void resetOptions() {
        // Partition
        m_optiPortion = 0.7;
        m_trainPortion = 0.7;

        // KNN
        m_kMax = 5;
        m_optimizeK = false;
        m_kOpted = m_kMax;

        // PSO
        m_popSize = 100;
        m_maxIter = 100;
        m_psoW = 0.7298;
        m_psoC1 = 1.49618;
        m_psoC2 = 1.49618;
        m_psoPhi = 0.1;

        m_initTypeString = "default";
        m_initPortion = new double[] { 1. };
        m_initProb = new double[] { 0.5 };

        m_pso = new PSOSearch();
    }

    private void logParameters(final String testType) {
        TideIOUtils.saveLog(m_logFile, "Dataset: " + m_dataFile);
        TideIOUtils.saveLog(m_logFile, "# of instances:   " + m_all.numInstances());
        TideIOUtils.saveLog(m_logFile, "# of attributes:  " + (m_all.numAttributes() - 1));
        TideIOUtils.saveLog(m_logFile, "# of classes:     " + m_all.numClasses());
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, "Opti Size: " + df2.format(m_optiPortion) + " (" + m_nOpti + " instances)");
        TideIOUtils.saveLog(m_logFile, "Train Size: " + df2.format(m_trainPortion) + " (" + m_nTrain + " instances)");
        TideIOUtils.saveLog(m_logFile, "Test Size:  " + df2.format(1 - m_trainPortion) + " (" + (m_nInstances - m_nTrain) + " instances)");
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, "CV Folds: " + m_nFoldCV);
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, "kNN Max k: " + m_kMax);
        TideIOUtils.saveLog(m_logFile, "kNN Optimize: " + (m_optimizeK ? "true" : "false"));
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, "PopSize: " + m_popSize);
        TideIOUtils.saveLog(m_logFile, "MaxIter: " + m_maxIter);
        TideIOUtils.saveLog(m_logFile, "psoW: " + m_psoW);
        TideIOUtils.saveLog(m_logFile, "psoC1: " + m_psoC1);
        TideIOUtils.saveLog(m_logFile, "psoC2: " + m_psoC2);
        TideIOUtils.saveLog(m_logFile, "psoPhi: " + m_psoPhi);
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, "Init Types: " + m_initTypeString);
        TideIOUtils.saveLog(m_logFile, "Init Portions: " + Arrays.toString(m_initPortion));
        TideIOUtils.saveLog(m_logFile, "Init Probs: " + Arrays.toString(m_initProb));
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, "RandSeed Shuffle: " + m_randSeedShuffle);
        TideIOUtils.saveLog(m_logFile, "RandSeed Algo: " + m_randSeedAlgorithm);
        TideIOUtils.saveLog(m_logFile, "");

        TideIOUtils.saveLog(m_logFile, m_kOpted + "-NN Error (Baseline): " + m_errorKnn);

        TideIOUtils.saveLog(m_logFile, "Test Type: " + testType);
        TideIOUtils.saveLog(m_logFile, "");
    }

    private void readData() throws FileNotFoundException, IOException {
        m_all = new Instances(new BufferedReader(new FileReader(m_dataFile)));

        m_all.setClassIndex(m_all.numAttributes() - 1);

        m_nInstances = m_all.numInstances();
        m_nOpti = (int) Math.floor(m_nInstances * m_optiPortion);
        m_nTrain = (int) Math.floor(m_nInstances * m_trainPortion);
    }

    private void prepData() {
        Random rnd = new Random(m_randSeedShuffle);
        m_all.randomize(rnd);
        TideFSUtils.distributeInstancesByClassesEvenly(m_all);

        m_opti = new Instances(m_all, 0, m_nOpti);
        m_train = new Instances(m_all, 0, m_nTrain);
        m_test = new Instances(m_all, m_nTrain, m_nInstances - m_nTrain);

        m_randSeedAlgorithm = rnd.nextLong();
    }

    private void initKnn() throws Exception {
        m_ibk = new IBk();
        m_ibk.setKNN(m_kMax);
        m_ibk.setCrossValidate(m_optimizeK);
        m_ibk.buildClassifier(m_train);
        m_errorKnn = TideFSUtils.errorRate(m_ibk, m_test);
        m_kOpted = m_ibk.getKNN();
        m_ibk.setCrossValidate(false);
        m_eva = new Evaluator(m_ibk, m_opti, m_nFoldCV, m_randSeedAlgorithm);
    }

    public void setPopSize(final int popSize) {
        m_popSize = popSize;
    }

    public int getPopSize() {
        return m_popSize;
    }

    public void setMaxIter(final int maxIter) {
        m_maxIter = maxIter;
    }

    public int getMaxIter() {
        return m_maxIter;
    }
}
