package weka.attributeSelection;

import java.io.Serializable;
import java.util.Random;

import weka.core.RevisionHandler;
import weka.core.RevisionUtils;

public class PSOParticle implements Cloneable, Serializable, RevisionHandler {

    private static final long serialVersionUID = 4485423629556221056L;

    private long              m_randSeed       = 0;

    private int               m_size;

    private double[]          m_x;
    private double[]          m_v;

    private double            m_objective      = -Double.MAX_VALUE;
    private double            m_fitness        = -Double.MAX_VALUE;

    private double            m_min            = 0;
    private double            m_max            = 1;

    public PSOParticle(PSOParticle p) {
        m_randSeed = p.getRandSeed();
        m_size = p.getSize();
        setPositions(p.getPositions());
        setVelocities(p.getVelocities());
        m_objective = p.getObjective();
        m_fitness = p.getFitness();
        setRange(p.getRange());
    }

    public PSOParticle(int size) {
        m_size = size;
        m_x = new double[m_size];
        m_v = new double[m_size];
    }

    public void update(double w, double c1, double r1, PSOParticle gbest, double c2, double r2, PSOParticle pbest) {
        assert (m_size == gbest.getSize() && m_size == pbest.getSize());

        double[] x1 = gbest.getPositions();
        double[] x2 = pbest.getPositions();

        for (int i = 0; i < m_size; i++) {
            m_v[i] = w * m_v[i] + c1 * r1 * (x1[i] - m_x[i]) + c2 * r2 * (x2[i] - m_x[i]);
        }
        updatePosition();
    }

    public void updateCSO(double r1, double r2, double r3, double phi, double[] winnerPos, double[] meanPos) {
        assert (m_size == winnerPos.length && m_size == meanPos.length);

        for (int i = 0; i < m_size; i++) {
            m_v[i] = r1 * m_v[i] + r2 * (winnerPos[i] - m_x[i]) + phi * r3 * (meanPos[i] - m_x[i]);
        }
        updatePosition();
    }

    private void updatePosition() {
        for (int i = 0; i < m_size; i++) {
            m_x[i] += m_v[i];
        }
        checkRange();
    }

    public void init() {
        Random rnd = new Random(m_randSeed);
        for (int i = 0; i < m_size; i++) {
            m_x[i] = rnd.nextDouble() * (m_max - m_min) + m_min;
        }
    }

    public double[] getRange() {
        double[] ret = new double[2];
        ret[0] = m_min;
        ret[1] = m_max;
        return ret;
    }

    public double getRangeMin() {
        return m_min;
    }

    public double getRangeMax() {
        return m_max;
    }

    public void setRange(double[] range) {
        assert (range.length == 2);
        setRange(range[0], range[1]);
    }

    public void setRange(double min, double max) {
        m_min = min;
        m_max = max;
    }

    public void setRangeMin(double min) {
        m_min = min;
    }

    public void setRangeMax(double max) {
        m_max = max;
    }

    private void checkRange() {
        for (int i = 0; i < m_size; i++) {
            checkRange(i);
        }
    }

    private void checkRange(int index) {
        if (m_x[index] < m_min) m_x[index] = m_min;
        if (m_x[index] > m_max) m_x[index] = m_max;
    }

    public long getRandSeed() {
        return m_randSeed;
    }

    public void setRandSeed(long seed) {
        m_randSeed = seed;
    }

    public double getObjective() {
        return m_objective;
    }

    public void setObjective(double objective) {
        m_objective = objective;
    }

    public double getFitness() {
        return m_fitness;
    }

    public void setFitness(double fitness) {
        m_fitness = fitness;
    }

    public double[] getPositions() {
        return m_x.clone();
    }

    public void setPositions(double[] positions) {
        m_x = positions.clone();
        checkRange();
    }

    public double[] getVelocities() {
        return m_v.clone();
    }

    public void setVelocities(double[] velocities) {
        m_v = velocities.clone();
        checkRange();
    }

    public double getPosition(int index) {
        return m_x[index];
    }

    public void setPosition(int index, double value) {
        m_x[index] = value;
        checkRange(index);
    }

    public double getVelocity(int index) {
        return m_v[index];
    }

    public void setVelocity(int index, double value) {
        m_v[index] = value;
        checkRange(index);
    }

    public int getSize() {
        return this.m_size;
    }

    public String toString() {
        StringBuffer result = new StringBuffer();

        result.append(m_size + ": {");

        for (int i = 0; i < m_size - 1; i++) {
            result.append(String.format("(%.4f, %.4f), ", m_x[i], m_v[i]));
        }

        result.append(String.format("(%.4f, %.4f)}", m_x[m_size - 1], m_v[m_size - 1]));

        return result.toString();
    }

    public Object clone() throws CloneNotSupportedException {
        PSOParticle tmp = new PSOParticle(m_size);
        tmp.m_randSeed = m_randSeed;
        tmp.m_size = m_size;
        tmp.m_x = m_x.clone();
        tmp.m_v = m_v.clone();
        tmp.m_objective = m_objective;
        tmp.m_fitness = m_fitness;
        tmp.m_min = m_min;
        tmp.m_max = m_max;
        return tmp;
    }

    @Override
    public String getRevision() {
        return RevisionUtils.extract("$Revision: 1 $");
    }

}
