package weka.attributeSelection;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Random;
import java.util.concurrent.TimeUnit;

import weka.core.Instances;
import weka.core.Utils;

public class PSOSearch extends ASSearch {

    private static final long         serialVersionUID   = 2485400505626877928L;

    private static final long         minutesPerLogSaves = 1;

    private boolean                   m_hasClass;
    private int                       m_classIndex;

    private int                       m_numAttribs;

    private PSOParticle[]             m_pop;
    private int                       m_popSize;

    private PSOParticle               m_gBest;
    private PSOParticle[]             m_pBest;

    private Hashtable<BitSet, Double> m_lookupTable;
    private Hashtable<BitSet, Double> m_oldLookupTable;

    private String                    m_lookupTableFilePath;
    private long                      m_lastSaveTime;

    private int[]                     m_numEvaluation;
    private double[]                  m_histFitness;
    private int[]                     m_histNumSelected;
    private List<List<Integer>>       m_histSelected;

    private Random                    m_random;
    private long                      m_seed;

    private double[]                  m_instProb;
    private double[]                  m_instPortion;

    private double                    m_w;                                      // inertia weight
    private double                    m_c1;                                     // cognitive parameter
    private double                    m_c2;                                     // social parameter

    private double                    m_phi;                                    // CSO phi

    private double                    m_threshold;                              // threshold of a feature should be selected

    private double                    m_sumFitness;

    private double                    m_maxFitness;
    private double                    m_minFitness;
    private double                    m_avgFitness;

    private int                       m_maxIter;

    public PSOSearch() {
        resetOptions();
    }

    @Override
    public int[] search(ASEvaluation ASEval, Instances data) throws Exception {
        return search(ASEval, data, 0);
    }

    public int[] search(ASEvaluation ASEval, Instances data, int algo) throws Exception {
        m_gBest = null;

        if (!(ASEval instanceof SubsetEvaluator)) {
            throw new Exception(ASEval.getClass().getName() + " is not a Subset evaluator!");
        }

        if (ASEval instanceof UnsupervisedSubsetEvaluator) {
            m_hasClass = false;
        } else {
            m_hasClass = true;
            m_classIndex = data.classIndex();
        }

        SubsetEvaluator ASEvaluator = (SubsetEvaluator) ASEval;
        m_numAttribs = data.numAttributes();

        m_lookupTable = new Hashtable<BitSet, Double>(m_popSize * m_maxIter);
        readLookupTable();
        if (m_oldLookupTable == null) {
            m_oldLookupTable = new Hashtable<BitSet, Double>(m_popSize * m_maxIter);
        }

        m_pop = new PSOParticle[m_popSize];
        m_pBest = new PSOParticle[m_popSize];

        m_numEvaluation = new int[m_maxIter + 1];
        m_histFitness = new double[m_maxIter + 1];
        m_histNumSelected = new int[m_maxIter + 1];
        m_histSelected = new ArrayList<List<Integer>>(m_maxIter + 1);

        initPopulation();
        evaluatePopulation(ASEvaluator, 0);
        populationStatistics();
        scalePopulation();
        updateBest(0);

        int i = 1;

        while (i <= m_maxIter) {
            System.out.print(".");
            if (i % 100 == 0 || i == m_maxIter) System.out.println(" " + i + "/" + m_maxIter);
            generation(algo);
            evaluatePopulation(ASEvaluator, i);
            populationStatistics();
            scalePopulation();
            updateBest(i);
            i++;
        }

        return attributeList(m_gBest);
    }

    private BitSet particle2bitset(PSOParticle p) {
        assert (m_numAttribs == p.getSize());

        BitSet bits = new BitSet(m_numAttribs);

        double mid = (p.getRangeMax() - p.getRangeMin()) * m_threshold + p.getRangeMin();

        double[] selection = p.getPositions();

        for (int i = 0; i < m_numAttribs; i++) {
            if (selection[i] > mid) {
                bits.set(i, true);
            } else {
                bits.set(i, false);
            }
        }

        return bits;
    }

    private int[] attributeList(PSOParticle p) {
        BitSet group = particle2bitset(p);

        int count = 0;

        for (int i = 0; i < m_numAttribs; i++) {
            if (group.get(i)) {
                count++;
            }
        }

        int[] list = new int[count];
        count = 0;

        for (int i = 0; i < m_numAttribs; i++) {
            if (group.get(i)) {
                list[count] = i;
                count++;
            }
        }

        return list;
    }

    private void initPopulation() throws Exception {
        int[] portion = new int[m_instPortion.length];
        double accumulatePortion = 0;
        for (int i = 0; i < m_instPortion.length; i++) {
            accumulatePortion += m_instPortion[i];
            portion[i] = (int) Math.round(m_popSize * accumulatePortion);
        }
        int portionId = 0;
        for (int i = 0; i < m_popSize; i++) {
            if (i > portion[portionId]) {
                portionId++;
            }
            m_pop[i] = new PSOParticle(m_numAttribs);
            m_pop[i].setRange(0, m_threshold / (1 - m_instProb[portionId]) + Double.MIN_NORMAL);
            m_pop[i].setRandSeed(getRandom().nextLong());
            m_pop[i].init();
            m_pop[i].setRange(0, 1);
            if (m_hasClass) {
                m_pop[i].setPosition(m_classIndex, m_pop[i].getRangeMax());
                m_pop[i].setVelocity(m_classIndex, m_pop[i].getRangeMax());
            }
        }
    }

    private void evaluatePopulation(SubsetEvaluator ASEvaluator, int currIter) throws Exception {

        BitSet auxBitSet;

        int count = 0;

        for (int i = 0; i < m_popSize; i++) {
            auxBitSet = particle2bitset(m_pop[i]);
            if (m_lookupTable.containsKey(auxBitSet) == false) {
                if (m_oldLookupTable != null && m_oldLookupTable.containsKey(auxBitSet) == true) {
                    m_pop[i].setObjective(m_oldLookupTable.get(auxBitSet));
                } else {
                    m_pop[i].setObjective(ASEvaluator.evaluateSubset(auxBitSet));
                }
                m_lookupTable.put((BitSet) auxBitSet.clone(), m_pop[i].getObjective());
                m_oldLookupTable.put((BitSet) auxBitSet.clone(), m_pop[i].getObjective());
                count++;
            } else {
                m_pop[i].setObjective(m_lookupTable.get(auxBitSet));
            }
        }
        m_numEvaluation[currIter] = count;

        long currSystemTime = System.nanoTime();
        if ((currIter == m_maxIter) || (TimeUnit.NANOSECONDS.toMinutes(currSystemTime - m_lastSaveTime) >= minutesPerLogSaves)) {
            saveLookupTable();
            m_lastSaveTime = currSystemTime;
        }
    }

    private void populationStatistics() {
        m_sumFitness = m_minFitness = m_maxFitness = m_pop[0].getObjective();

        for (int i = 1; i < m_popSize; i++) {
            m_sumFitness += m_pop[i].getObjective();
            if (m_pop[i].getObjective() > m_maxFitness) {
                m_maxFitness = m_pop[i].getObjective();
            } else if (m_pop[i].getObjective() < m_minFitness) {
                m_minFitness = m_pop[i].getObjective();
            }
        }
        m_avgFitness = (m_sumFitness / m_popSize);
    }

    private void scalePopulation() {
        int j;
        double a = 0;
        double b = 0;
        double fmultiple = 2.0;
        double delta;

        // pre-scale
        if (m_minFitness > ((fmultiple * m_avgFitness - m_maxFitness) / (fmultiple - 1.0))) {
            delta = m_maxFitness - m_avgFitness;
            a = ((fmultiple - 1.0) * m_avgFitness / delta);
            b = m_avgFitness * (m_maxFitness - fmultiple * m_avgFitness) / delta;
        } else {
            delta = m_avgFitness - m_minFitness;
            a = m_avgFitness / delta;
            b = -m_minFitness * m_avgFitness / delta;
        }

        // scale pop
        m_sumFitness = 0;
        for (j = 0; j < m_popSize; j++) {
            if (a == Double.POSITIVE_INFINITY || a == Double.NEGATIVE_INFINITY || b == Double.POSITIVE_INFINITY || b == Double.NEGATIVE_INFINITY) {
                m_pop[j].setFitness(m_pop[j].getObjective());
            } else {
                m_pop[j].setFitness(Math.abs((a * m_pop[j].getObjective() + b)));
            }
            m_sumFitness += m_pop[j].getFitness();
        }
    } // scalePopulation

    private void shufflePopulation() {
        for (int i = m_popSize - 1; i > 0; i--) {
            int idx = getRandom().nextInt(i + 1);
            PSOParticle tmp = m_pop[idx];
            m_pop[idx] = m_pop[i];
            m_pop[i] = tmp;
        }
    }

    private void updateBest(int currIter) throws Exception {
        updateParticleBest();
        updateGlobalBest(currIter);
    }

    private void updateParticleBest() throws Exception {
        double obj1, obj2;
        int n1, n2;

        if (m_pBest[0] == null) {
            for (int i = 0; i < m_popSize; i++) {
                m_pBest[i] = (PSOParticle) m_pop[i].clone();
            }
        } else {
            for (int i = 0; i < m_popSize; i++) {
                PSOParticle current = m_pop[i];
                PSOParticle pBest = m_pBest[i];

                obj1 = current.getObjective();
                obj2 = pBest.getObjective();

                n1 = countFeatures(current);
                n2 = countFeatures(pBest);

                if ((obj1 > obj2) || (Utils.eq(obj1, obj2) && (n1 < n2))) {
                    m_pBest[i] = (PSOParticle) current.clone();
                }
            }
        }
    }

    private void updateGlobalBest(int currIter) throws Exception {

        PSOParticle cBest;

        if (m_gBest == null) {
            cBest = (PSOParticle) m_pBest[0].clone();
        } else {
            cBest = m_gBest;
        }

        double obj, objBest = cBest.getObjective();
        int n, nBest = countFeatures(cBest);

        if (m_maxFitness - m_minFitness > 0) {
            for (int i = 1; i < m_popSize; i++) {
                obj = m_pBest[i].getObjective();
                n = countFeatures(m_pBest[i]);

                if ((obj > objBest) || (Utils.eq(obj, objBest) && (n < nBest))) {
                    cBest = m_pBest[i];
                    objBest = obj;
                    nBest = n;
                }
            }
        } else {
            nBest = m_numAttribs;
            for (int i = 1; i < m_popSize; i++) {
                n = countFeatures(m_pBest[i]);
                if (n < nBest) {
                    cBest = m_pBest[i];
                    nBest = n;
                }
            }
        }

        m_gBest = (PSOParticle) cBest.clone();

        m_histFitness[currIter] = -m_gBest.getObjective();
        m_histNumSelected[currIter] = countFeatures(m_gBest);
        m_histSelected.add(getFeatures(m_gBest));
    }

    private int countFeatures(PSOParticle p) {
        BitSet featureSet = particle2bitset(p);
        int count = 0;
        for (int i = 0; i < m_numAttribs; i++) {
            if (featureSet.get(i)) {
                count++;
            }
        }
        return count;
    }

    private List<Integer> getFeatures(PSOParticle p) {
        List<Integer> ret = new ArrayList<Integer>(countFeatures(p));

        int[] list = attributeList(p);

        for (int i : list) {
            ret.add(i);
        }

        return ret;
    }

    private Random getRandom() {
        if (m_random == null) {
            m_random = new Random(m_seed);
        }
        return m_random;
    }

    private void generation(int algo) throws Exception {
        int halfPopSize = m_popSize / 2;

        double obj1, obj2;
        int n1, n2;
        int idxLoser, idxWinner;

        double[] currPos;
        double[] meanPos = new double[m_numAttribs];

        switch (algo) {
            case 0:
                for (int i = 0; i < m_popSize; i++) {
                    m_pop[i].update(m_w, m_c1, getRandom().nextDouble(), m_gBest, m_c2, getRandom().nextDouble(), m_pBest[i]);

                }
                break;
            case 1:
                shufflePopulation();
                for (int i = 0; i < m_popSize; i++) {
                    currPos = m_pop[i].getPositions();
                    for (int j = 0; j < m_numAttribs; j++) {
                        meanPos[j] += currPos[j];
                    }
                }
                for (int j = 0; j < m_numAttribs; j++) {
                    meanPos[j] /= m_numAttribs;
                }
                for (int i = 0; i < m_popSize / 2; i++) {
                    idxLoser = i;
                    idxWinner = i + halfPopSize;

                    obj1 = m_pop[i].getObjective();
                    obj2 = m_pop[i + halfPopSize].getObjective();

                    n1 = countFeatures(m_pop[i]);
                    n2 = countFeatures(m_pop[i + halfPopSize]);

                    if ((obj1 > obj2) || (Utils.eq(obj1, obj2) && (n1 < n2))) {
                        idxLoser = i + halfPopSize;
                        idxWinner = i;
                    }
                    m_pop[idxLoser].updateCSO(getRandom().nextDouble(), getRandom().nextDouble(), getRandom().nextDouble(), m_phi,
                                              m_pop[idxWinner].getPositions(), meanPos);
                }
                break;
        }
    }

    public int[] getNumEvaluation() {
        return m_numEvaluation;
    }

    public double[] getHistFitness() {
        return m_histFitness;
    }

    public int[] getHistNumSelected() {
        return m_histNumSelected;
    }

    public List<List<Integer>> getHistSelected() {
        return m_histSelected;
    }

    private void resetOptions() {
        m_pop = null;
        m_popSize = 20;
        m_maxIter = 20;
        m_seed = 1;

        m_instPortion = new double[] { 1.0 };
        m_instProb = new double[] { 0.5 };

        m_w = 1;
        m_c1 = 2;
        m_c2 = 2;
        m_phi = 0.1;
        m_threshold = 0.5;
    }

    public void setPopulationSize(int p) {
        m_popSize = p;
    }

    public int getPopulationSize() {
        return m_popSize;
    }

    public void setIterations(int iterations) {
        m_maxIter = iterations;
    }

    public int getIterations() {
        return m_maxIter;
    }

    public void setSeed(long s) {
        m_seed = s;
        m_random = new Random(m_seed);
    }

    public long getSeed() {
        return m_seed;
    }

    public void setInstPortionProb(double[] portion, double[] prob) {
        assert (Utils.eq(Arrays.stream(portion).sum(), 1));
        assert (portion.length == prob.length);
        m_instPortion = Arrays.copyOf(portion, portion.length);
        m_instProb = Arrays.copyOf(prob, prob.length);
    }

    public double[] getInstPortion() {
        return Arrays.copyOf(m_instPortion, m_instPortion.length);
    }

    public double[] getInstProb() {
        return Arrays.copyOf(m_instProb, m_instProb.length);
    }

    public void setW(double w) {
        m_w = w;
    }

    public double getW() {
        return m_w;
    }

    public void setC1(double c1) {
        m_c1 = c1;
    }

    public double getC1() {
        return m_c1;
    }

    public void setC2(double c2) {
        m_c2 = c2;
    }

    public double getC2() {
        return m_c2;
    }

    public void setPhi(double phi) {
        m_phi = phi;
    }

    public double getPhi() {
        return m_phi;
    }

    public void setThreshold(double threshold) {
        assert (threshold > 0 && threshold < 1);
        m_threshold = threshold;
    }

    public double getThreshold() {
        return m_threshold;
    }

    public void setLookupTableFilePath(String path) {
        m_lookupTableFilePath = path;
    }

    private void saveLookupTable() throws IOException {
        FileOutputStream fos = null;
        ObjectOutputStream oos = null;
        if (m_lookupTableFilePath != null) {
            try {
                File f = new File(m_lookupTableFilePath);
                if (f.exists() && !f.isDirectory()) {
                    Files.copy(Paths.get(m_lookupTableFilePath), Paths.get(m_lookupTableFilePath + ".bak"), StandardCopyOption.REPLACE_EXISTING);
                }
                fos = new FileOutputStream(m_lookupTableFilePath);
                oos = new ObjectOutputStream(fos);
                oos.writeObject(m_oldLookupTable);
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                if (oos != null) {
                    oos.close();
                }
                if (fos != null) {
                    fos.close();
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void readLookupTable() throws IOException {
        FileInputStream fis = null;
        ObjectInputStream ois = null;

        if (m_lookupTableFilePath != null) {
            File objFile = new File(m_lookupTableFilePath);
            if (objFile.exists() && !objFile.isDirectory()) {
                try {
                    fis = new FileInputStream(m_lookupTableFilePath);
                    ois = new ObjectInputStream(fis);
                    m_oldLookupTable = (Hashtable<BitSet, Double>) ois.readObject();
                } catch (Exception e) {
                    e.printStackTrace();
                } finally {
                    if (ois != null) {
                        ois.close();
                    }
                    if (fis != null) {
                        fis.close();
                    }
                }
            } else {
                m_oldLookupTable = null;
            }
        } else {
            m_oldLookupTable = null;
        }
    }
}
